import asyncio
import unittest
from gizmo import BaseGremlinClient, AsyncGremlinClient, GremlinClient


class BaseGremlinClientTests(unittest.TestCase):

    def setUp(self):
        self.client = BaseGremlinClient("ws://localhost:8182/")

    def test_connection(self):
        @asyncio.coroutine
        def conn_coro():
            conn = yield from self.client.connect()
            self.assertTrue(conn.open)
        self.client.run_until_complete(conn_coro())


class AsyncGremlinClientTests(unittest.TestCase):

    def setUp(self):
        self.client = AsyncGremlinClient("ws://localhost:8182/")

    def test_01_connection(self):
        @asyncio.coroutine
        def conn_coro():
            conn = yield from self.client.connect()
            self.assertTrue(conn.open)
        self.client.run_until_complete(conn_coro())

    ### Basic server communication tests. This implements the opening example
    ### from http://www.tinkerpop.com/docs/3.0.0.M7/###

    def test_02_graph_create(self):
        @asyncio.coroutine
        def graph_open_coro():
            yield from self.client.task(self.client.send_receive,
                                        "g = TinkerGraph.open()")
            f = yield from self.client.messages.get()
            self.assertEqual(f.status_code, 200)
        self.client.run_until_complete(graph_open_coro())

    def test_03_node_edge_create(self):
        @asyncio.coroutine
        def node_edge_create_coro():
            yield from self.client.task(
                self.client.send_receive,
                "g.V().remove(); g.E().remove();",
                collect=False
            )

            yield from self.client.task(
                self.client.send_receive,
                ("gremlin = g.addVertex(label,'software','name','gremlin');" +
                 "gremlin.property('created', 2009);" +
                 "blueprints = g.addVertex(label,'software','name','blueprints');" +
                 "gremlin.addEdge('dependsOn',blueprints);" +
                 "blueprints.property('created',2010);" +
                 "blueprints.property('created').remove()"),
                collect=False)

            yield from self.client.task(
                self.client.send_receive,
                "g.V().count()")
            f = yield from self.client.messages.get()
            self.assertEqual(f[0], 2)

            yield from self.client.task(
                self.client.send_receive,
                "g.E().count()")
            f = yield from self.client.messages.get()
            self.assertEqual(f[0], 1)

            yield from self.client.task(
                self.client.send_receive,
                "g.V()")
            nodes = ["gremlin", "blueprints"]
            f = yield from self.client.messages.get()
            label = f[0]["label"]
            self.assertEqual(label, "software")
            name1 = f[0]["name"][0]
            self.assertTrue(name1 in nodes)
            nodes.remove(name1)
            print("Successfully created node of type software: gremlin")
            name2 = f[1]["name"][0]
            self.assertTrue(name2 in nodes)
            crtd = f[1].get("created", "")
            self.assertEqual(crtd, "")
            print("Successfully created node of type software: blueprints")

            yield from self.client.task(
                self.client.send_receive,
                "g.E()")
            f = yield from self.client.messages.get()
            label = f[0]["label"]
            in_label = f[0]["inVLabel"]
            out_label = f[0]["outVLabel"]
            self.assertEqual(label, "dependsOn")
            self.assertEqual(in_label, "software")
            self.assertEqual(out_label, "software")
            print("Successfully created edge of type dependsOn: " +
                  "software dependsOn software")
        self.client.run_until_complete(node_edge_create_coro())

    ### Tests for the consumer - if returns value, add to the message queue.

    def test_04_consumer_return(self):
        @asyncio.coroutine
        def consumer_return_coro():
            yield from self.client.task(
                self.client.send_receive,
                "1 + 1",
                consumer=lambda x: x[0] ** 2)
            while not self.client.messages.empty():
                f = yield from self.client.messages.get()
                self.assertEqual(f, 4)
                print("Consumer added results to messages")
        self.client.run_until_complete(consumer_return_coro())

    def test_05_consumer_noreturn(self):
        @asyncio.coroutine
        def consumer_noreturn_coro():
            yield from self.client.task(
                self.client.send_receive,
                "1 + 1",
                consumer=lambda x: print(x[0] ** 2))
            self.assertTrue(self.client.messages.empty())
            print("Consumer did not add results to messages")
        self.client.run_until_complete(consumer_noreturn_coro())

    ### Collect false turns off collection.

    def test_06_collect_false(self):
        @asyncio.coroutine
        def consumer_return_coro():
            yield from self.client.task(
                self.client.send_receive,
                "1 + 1",
                consumer=lambda x: x[0] ** 2,
                collect=False)
            self.assertTrue(self.client.messages.empty())
            print("No collection performed")
        self.client.run_until_complete(consumer_return_coro())

    ### Run async parallel tasks.

    @asyncio.coroutine
    def slowjson(self, consumer=None):
        if consumer is None:
            consumer = lambda x: x[0]
        yield from asyncio.sleep(0.25)
        yield from self.client.task(
            self.client.send_receive,
            "g.V().has(n, val).values(n)",
            bindings={"n": "name", "val": "gremlin"},
            consumer=consumer)

    def test_07_parallel_tasks(self):
        task1 = self.client.add_task(self.slowjson)
        task2 = self.client.add_task(
            self.client.send_receive,
            "g.V().has(n, val).values(n)",
            bindings={"n": "name", "val": "blueprints"},
            consumer=lambda x: x[0])
        self.client.run_tasks()
        names = []
        @asyncio.coroutine
        def parallel_task_coro():
            while not self.client.messages.empty():
                f = yield from self.client.messages.get()
                names.append(f)
        self.client.run_until_complete(parallel_task_coro())

        self.assertEqual(names[0], "blueprints")
        self.assertEqual(names[1], "gremlin")
        print("Asynchronous task execution successful.")

    ### Enqueue/dequeue tests.

    def test_08_enqueue_dequeue(self):
        @asyncio.coroutine
        def enqueue_dequeue_coro():
            yield from self.client.enqueue_task(self.slowjson)
            yield from self.client.enqueue_task(
                self.client.send_receive,
                "g.V().has(n, val).values(n)",
                bindings={"n": "name", "val": "blueprints"},
                consumer=lambda x: x[0])
            yield from self.client.dequeue_task()
            yield from self.client.dequeue_task()
            mssg1 = yield from self.client.messages.get()
            self.assertEqual(mssg1, "gremlin")
            print("Successfully dequeued slow operation: gremlin")
            mssg2 = yield from self.client.messages.get()
            self.assertEqual(mssg2, "blueprints")
            print("Successfully dequeued fast operation: blueprints")

        self.client.run_until_complete(enqueue_dequeue_coro())

    def test_09_dequeue_all(self):
        messages = []
        @asyncio.coroutine
        def dequeue_all_coro():
            yield from self.client.enqueue_task(
                self.slowjson,
                consumer=lambda x: x[0])
            yield from self.client.enqueue_task(
                self.client.send_receive,
                "g.V().has(n, val).values(n)",
                bindings={"n": "name", "val": "blueprints"},
                consumer=lambda x: x[0])
            yield from self.client.dequeue_all()
            while not self.client.messages.empty():
                mssg = yield from self.client.messages.get()
                messages.append(mssg)
            self.assertEqual(len(messages), 2)
            self.assertEqual(messages[0], "gremlin")
            self.assertEqual(messages[1], "blueprints")
            print("Successfully deququeued all.")
        self.client.run_until_complete(dequeue_all_coro())

    def test_10_async_dequeue_all(self):

        @asyncio.coroutine
        def check_messages_coro():
            mssg1 = yield from self.client.messages.get()
            mssg2 = yield from self.client.messages.get()
            self.assertEqual(mssg1, "blueprints")
            print("Async dequeued blueprints")
            self.assertEqual(mssg2, "gremlin")
            print("Async dequeued gremlin")

        @asyncio.coroutine
        def async_dequeue_consumer(q):
            coro, args, kwargs = yield from q.get()
            task = self.client.task(coro, *args, **kwargs)
            f = yield from task

        @asyncio.coroutine
        def enqueue_all_coro():
            yield from self.client.enqueue_task(self.slowjson,
                consumer=lambda x: x[0])
            yield from self.client.enqueue_task(
                self.client.send_receive,
                "g.V().has(n, val).values(n)",
                bindings={"n": "name", "val": "blueprints"},
                consumer=lambda x: x[0])
        self.client.run_until_complete(enqueue_all_coro())
        self.client.async_dequeue_all(async_dequeue_consumer)
        self.client.run_until_complete(check_messages_coro())


class GremlinClientTests(unittest.TestCase):

    def setUp(self):
        self.client = GremlinClient("ws://localhost:8182/")

    def test_connection(self):
        @asyncio.coroutine
        def conn_coro():
            conn = yield from self.client.connect()
            self.assertTrue(conn.open)
        self.client.run_until_complete(conn_coro())

    def test_execute(self):
        self.client.execute(
            "g.V().has(n, val).values(n)",
            bindings={"n": "name", "val": "blueprints"},
            consumer=lambda x: x[0])
        self.assertEqual(self.client._messages[0], "blueprints")
        self.client.execute(
            "g.V().has(n, val).values(n)",
            bindings={"n": "name", "val": "gremlin"},
            consumer=lambda x: x[0])
        self.assertEqual(self.client._messages[1], "gremlin")
        for x in self.client:
            print("Retrieved: {}".format(x))

if __name__ == "__main__":
    unittest.main()



class GremlinResponse(list):
    """
    Parse and flatten the Gremlin Server's response a bit. Make standard use
    case easier for end user to process.
    """
    def __init__(self, message):
        super().__init__()
        data = message["result"].get("data", "")
        if data:
            for datum in data:
                if isinstance(datum, dict):
                    try:
                        datum = parse_struct(datum)
                    except (KeyError, IndexError):
                        pass
                self.append(datum)
        self.meta = message["result"]["meta"]
        self.request_id = message["requestId"]
        self.status_code = message["status"]["code"]
        self.message = message["status"]["message"]
        self.attrs = message["status"]["attributes"]


def parse_struct(struct):
    output = {}
    for k, v in struct.items():
        if k != "properties":
            output[k] = v
    # This looses part of the response. I'm not familiar yet with
    # how property ids and property properties work yet. May change.
    properties = {k: [val["value"] for val in v] for (k, v) in
        struct["properties"].items()}
    output.update(properties)
    return output



import asyncio
import json
import ssl
import uuid
import websockets
from .response import GremlinResponse


class BaseGremlinClient:

    def __init__(self, uri='ws://localhost:8182/', loop=None, ssl=None,
                 protocol=None, **kwargs):
        self.uri = uri
        # SLL is untested. Need to set up secure server and try it out.
        # Will look something like this.
        if ssl:
            protocol = protocol or ssl.PROTOCOL_TLSv1
            ssl_context = ssl.SSLContext(protocol)
            ssl_context.load_verify_locations(ssl)
            ssl_context.verify_mode = ssl.CERT_REQUIRED
            kwargs['ssl'] = ssl_context
        self._sock = asyncio.async(self.connect(**kwargs))
        self._errors = []
        self._loop = loop or asyncio.get_event_loop()

    def get_sock(self):
        return self._sock
    sock = property(get_sock)

    def get_errors(self):
        return self._errors
    errors = property(get_errors)

    @asyncio.coroutine
    def connect(self, **kwargs):
        websocket = yield from websockets.connect(self.uri, **kwargs)
        return websocket

    @asyncio.coroutine
    def send(self, gremlin, bindings=None, lang="gremlin-groovy", op="eval",
             processor=""):
        payload = {
            "requestId": str(uuid.uuid4()),
            "op": op,
            "processor": processor,
            "args":{
                "gremlin": gremlin,
                "bindings": bindings,
                "language":  lang
            }
        }
        try:
            websocket = yield from self.sock
            self._sock = websocket
        except TypeError:
            websocket = self.sock
        yield from websocket.send(json.dumps(payload))
        return websocket

    def receive(consumer=None, collect=True):
        raise NotImplementedError

    @asyncio.coroutine
    def send_receive(self, gremlin, bindings=None, lang="gremlin-groovy",
                     op="eval", processor="", consumer=None, collect=True):
        yield from self.send(gremlin, bindings=bindings, lang=lang, op=op,
                             processor=processor)
        yield from self.receive(consumer=consumer, collect=collect)

    def run_until_complete(self, func):
        self._loop.run_until_complete(func)


class AsyncGremlinClient(BaseGremlinClient):

    def __init__(self, uri='ws://localhost:8182/', loop=None, **kwargs):
        super().__init__(uri=uri, loop=loop, **kwargs)
        self.messages = asyncio.Queue()
        self._tasks = []
        self.task_queue = asyncio.Queue()

    def get_tasks(self):
        return self._tasks
    tasks = property(get_tasks)

    def task(self, coro, *args, **kwargs):
        return asyncio.async(coro(*args, **kwargs))

    def add_task(self, coro, *args, **kwargs):
        task = self.task(coro, *args, **kwargs)
        self._tasks.append(task)
        return task

    @asyncio.coroutine
    def enqueue_task(self, coro, *args, **kwargs):
        task = (coro, args, kwargs)
        yield from self.task_queue.put(task)

    @asyncio.coroutine
    def dequeue_task(self):
        if not self.task_queue.empty():
            coro, args, kwargs = yield from self.task_queue.get()
            task = self.task(coro, *args, **kwargs)
            return (yield from task)

    @asyncio.coroutine
    def dequeue_all(self):
        while not self.task_queue.empty():
            coro, args, kwargs = yield from self.task_queue.get()
            task = self.task(coro, *args, **kwargs)
            f = yield from task

    def async_dequeue_all(self, coro, *args, **kwargs):
        q = self.task_queue
        coros = [asyncio.async(coro(q, *args, **kwargs)) for i in
            range(q.qsize())]
        self.run_until_complete(asyncio.wait(coros))

    @asyncio.coroutine
    def receive(self, consumer=None, collect=True):
        websocket = self.sock
        while True:
            # Will need to handle error here if websocket no message has been
            # sent.
            message = yield from websocket.recv()
            message = json.loads(message)
            message = GremlinResponse(message)
            if message.status_code == 200:
                if consumer:
                    message = consumer(message)
                if message and collect:
                    yield from self.messages.put(message)
            elif message.status_code == 299:
                break
            else:
                # Error handler here.
                verbose = "Request {} failed with status code {}: {}".format(
                    message.request_id, message.status_code, message.message)
                self._errors.append(verbose)
                print(verbose)

    def run_tasks(self):
        self.run_until_complete(asyncio.wait(self.tasks))


class GremlinClient(BaseGremlinClient):

    def __init__(self, uri='ws://localhost:8182/', loop=None, **kwargs):
        super().__init__(uri=uri, loop=loop, **kwargs)
        self._messages = []
        self._message_number = 0

    def get_messages(self):
        for message in self._messages:
            yield message
    messages = property(get_messages)

    def __iter__(self):
        return iter(self._messages)

    @asyncio.coroutine
    def receive(self, consumer=None, collect=True):
        websocket = self.sock
        while True:
            # Will need to handle error here if websocket no message has been
            # sent.
            message = yield from websocket.recv()
            message = json.loads(message)
            message = GremlinResponse(message)
            if message.status_code == 200:
                if consumer:
                    message = consumer(message)
                if message and collect:
                    self._messages.append(message)
            elif message.status_code == 299:
                break
            else:
                # Error handler here.
                verbose = "Request {} failed with status code {}: {}".format(
                    message.request_id, message.satus_code, message.message)
                self._errors.append(verbose)
                print(verbose)

    def execute(self, gremlin, bindings=None, lang="gremlin-groovy", op="eval",
                processor="", consumer=None, collect=True):
        coro = self.send_receive(gremlin, bindings=bindings, lang=lang, op=op,
            processor=processor, consumer=consumer, collect=collect)
        self.run_until_complete(coro)
        return self
